/* 
  Copyright 2001 Andrew McNab mcnab@hep.man.ac.uk University of Manchester

  This file is part of GridSite.

  GridSite is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GridSite is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GridSite; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

/*------------------------------------------------------------------------*
 * For more information about GridSite: http://www.gridpp.ac.uk/gridsite/ *
 *------------------------------------------------------------------------*/

#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

/* rh62? */
#include <sys/time.h>

#include "gridsite.h"

#include "config.h"

int   ishttps = 0, aclstatus;
char *userdn = NULL, *docroot = NULL, *pathtranslated = NULL, 
     *requestdir = NULL, *requesturi = NULL, *cgiposted;

/* yes, this will break every time we change the GRID_ACL_ symbols ...*/
static char *aclsyms[] = { "none",
                           "read",
                           "list",
                           "write",
                           "admin",
                           NULL     };

int voidstrcmp(const void *p, const void *q)
{
  char **pp, **qq;
  
  pp = (char **) p;
  qq = (char **) q; 
  
  return strcmp(*pp, *qq);
}

char **SortDir(char *dirpath)
{
  char         **array;
  int            n, i;
  DIR           *listDIR;
  struct dirent *oneentry;
  
  listDIR = opendir(dirpath);
  if (listDIR == NULL) return NULL;

  n=0; 
  while ((oneentry = readdir(listDIR)) != NULL)
         if ((oneentry->d_name[0] != '.') && 
             (index(oneentry->d_name, '~') == NULL) &&
             (strcmp(oneentry->d_name, "core") != 0))  
           {
             printf("<!-- %d %s -->\n", n, oneentry->d_name);
             ++n;
           }

  rewinddir(listDIR);

  array = (char **) malloc(sizeof(char *) * (n + 1));

  i=0;
  while (i <= n)
     {
       oneentry = readdir(listDIR);
       
       if (oneentry == NULL) break;
       
       if ((oneentry->d_name[0] != '.') && 
           (index(oneentry->d_name, '~') == NULL) &&
           (strcmp(oneentry->d_name, "core") != 0)) 
         {
           array[i] = strdup(oneentry->d_name);
           printf("<!-- %d %s -->\n", i, array[i]);
           ++i;
         }
     } 
     
  array[i] = (char *) NULL;      

  closedir(listDIR);

  qsort((void *) array, (size_t) i, sizeof(char *), voidstrcmp);

  for (i=0; i < n; ++i) printf("<!-- %d %s -->\n", i, array[i]);

  return array;
}

int ConvertLevelSym(char *levelsym)
{
  int i;

  for (i=0; aclsyms[i] != NULL; ++i) 
   if (strcmp(levelsym, aclsyms[i]) == 0) return i;

  return 0;
}

char *GetFromCGI(char *name)
{
  char   *p, *namepattern, *valuestart, *returnvalue;
  int     c, i, j, n;  
  static  char *nullstring = "";
  size_t  size_needed;

  namepattern = malloc(strlen(name) + 3);
  sprintf(namepattern, "&%s=", name);
  
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
           "%s: looking for %s in POSTed data", 
           getenv("REMOTE_ADDR"), namepattern);
#endif

  p = strstr(cgiposted, namepattern);
  free(namepattern);
  if (p == NULL)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
           "%s: %s not present in POSTed data", getenv("REMOTE_ADDR"), name);
#endif
      return nullstring;
    }
     
  valuestart = &p[strlen(name) + 2];

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
           "%s: valuestart is %s in POSTed data", 
           getenv("REMOTE_ADDR"), valuestart);
#endif

  for (n=0; valuestart[n] != '&'; ++n) ;
  
  returnvalue = malloc(n + 1);
  
  j=0;
  
  for (i=0; i < n; ++i) 
     {
       if ((i < n - 2) && (valuestart[i] == '%')) /* url encoded as %HH */
         {
           returnvalue[j] = 0;
           
           if (isdigit(valuestart[i+1])) 
                 returnvalue[j] += 16 * (valuestart[i+1] - '0');
           else if (isalpha(valuestart[i+1])) 
                 returnvalue[j] += 16 * (10 + tolower(valuestart[i+1]) - 'a');
                         
           if (isdigit(valuestart[i+2])) 
                 returnvalue[j] += valuestart[i+2] - '0';
           else if (isalpha(valuestart[i+2])) 
                 returnvalue[j] += 10 + tolower(valuestart[i+2]) - 'a';
                                                   
           i = i + 2;
         }
       else if (valuestart[i] == '+') returnvalue[j] = ' ';
       else                           returnvalue[j] = valuestart[i];
       
       ++j;
     }

  returnvalue[j] = '\0';

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
       "%s: returning %s from GetFromCGI", getenv("REMOTE_ADDR"), returnvalue);
#endif

  return returnvalue;
}

char *ReadInPathTran()
{
  int     c;
  char   *rawpage, *p;
  size_t  rawsize;
  struct  stat buf;
  FILE   *fp;

  if (stat(pathtranslated, &buf) == -1) return NULL;

  rawsize = buf.st_size;
  rawpage = malloc(rawsize + (size_t) 1);

  fp = fopen(pathtranslated, "r");
  if (fp == NULL) return NULL;

  p = rawpage;
  while ((c = fgetc(fp)) != EOF)
       { 
         *p = c;
         ++p;
       }

  fclose(fp);
  
  *p = '\0';

  return rawpage;
}

FILE *DocRootfopen(char *filename, char *mode)
/* open file, where filename is relative to DOCUMENT_ROOT */
{
  char *fullfilename;
  FILE *fp;
  
  fullfilename = malloc(strlen(docroot) + 2 + strlen(filename));
  
  sprintf(fullfilename, "%s/%s", docroot, filename);
  
  fp = fopen(fullfilename, mode);
  
  free(fullfilename);
  
  return fp;
}

void CopyFile(char *destination, char *source)
{
  int    c;
  FILE *ifp, *ofp;

  ifp = fopen(source, "r");
  
  if (ifp != NULL)
    {
      ofp = fopen(destination, "w");
      if (ofp != NULL)
        {   
          while ((c = fgetc(ifp)) != EOF) fputc(c, ofp);    

          fclose(ofp);
        }
        
      fclose(ifp);
    }
}

int CatFile(char *filename)
{
  int    c;
  FILE *fp;

  fp = DocRootfopen(filename, "r");
  
  if (fp != NULL)
    {
      while ((c = fgetc(fp)) != EOF) putchar(c);    
      fclose(fp);
      return 1;
    }
  
  return 0;
}

int CatPathFile(char *pathname, char *filename)
{
  char  *pathfile; 
  int   c;
  FILE *fp;

  pathfile = malloc(strlen(pathname) + strlen(filename) + 2);
  sprintf(pathfile, "%s/%s", pathname, filename);

  fp = DocRootfopen(pathfile, "r");
  free(pathfile);
  
  if (fp != NULL)
    {
      while ((c = fgetc(fp)) != EOF) putchar(c);    
      fclose(fp);
      return 1;
    }
  
  return 0;
}

int CatHeadOrFoot(char *filename)
{
  int  found = 0;
  char *tmp_requestdir, *p;

  tmp_requestdir = strdup(requestdir);  

  while (!(found = CatPathFile(tmp_requestdir, filename)))
       {
         p = rindex(tmp_requestdir, '/');
         if (p == NULL) break;
         *p = '\0';
         p = rindex(tmp_requestdir, '/');
         if (p == NULL) break;
         p[1] = '\0';
       }
  
  free(tmp_requestdir);
  return found;
}

int UpdateFileLog(char *action, char *filename, int filesize, 
                                                 time_t now_t, long now_usec)
{
  int   c;
  char *logfilename, *tmplogfilename;
  FILE *oldfp, *newfp;

  logfilename    = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 8);
  tmplogfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 31);
  sprintf(logfilename, "%s%s.%s.log", docroot, requestdir, filename);
  sprintf(tmplogfilename, "%s%s.%s.%ld.%06ld.log", docroot, requestdir, 
                                                 filename, now_t, now_usec);

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
       "%s: logfilename=%s tmplogfilename=", getenv("REMOTE_ADDR"), 
       logfilename, tmplogfilename);
#endif

  newfp = fopen(tmplogfilename, "w");
  if (newfp == NULL) return 0;

  oldfp = fopen(logfilename, "r");
  if (oldfp != NULL)
    {
      while ((c = fgetc(oldfp)) != EOF) fputc(c, newfp);
      fclose(oldfp);
    }

  fprintf(newfp, "%s %ld %ld %d %s\n", 
                 action, filesize, now_t, now_usec, userdn);
  fclose(newfp);

  rename(tmplogfilename, logfilename);
  return 1;
}

int DisplayFileDate()
{
  int    c, ret = 0;
  char  *p, *logfilename, *requestdir, *filename, oneline[161], lastline[161],
         s[81];
  FILE  *fp;
  time_t file_t;
  struct tm *file_tm;

  requestdir = strdup(requesturi);
  
  filename = rindex(requestdir, '/');
  *filename = '\0';
  ++filename;
  if (*filename == '\0') filename = "index.html";

  logfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 8);

  sprintf(logfilename, "%s/%s/.%s.log", docroot, requestdir, filename);

  fp = fopen(logfilename, "r");
  if (fp != NULL)
    {
      while (fgets(oneline, sizeof(oneline), fp) != NULL)
           {
             p = rindex(oneline, '\n');
             if (p != NULL) *p = '\0';
             strcpy(lastline, oneline);
           }
     
      fclose(fp);

      sscanf(lastline, "%*s %*d %ld", &file_t);
      file_tm = localtime(&file_t);     
      strftime(s, sizeof(s), "%a %e %B %Y", file_tm); 

      printf("<hr>Last updated %s\n . 
              <a href=\"%s?history\">View page history</a>", s, requesturi);
      ret = 1;
    }

  free(requestdir);

  return ret;
}

char *GetMimeType(char *filename)
{
  int i, n, nn;

  n = strlen(filename);
  
  for (i=1; extensions[i] != NULL; i=i+2)
     {
       nn = strlen(extensions[i]);
        
       if (n > nn) 
        if (strcmp(&(filename[n-nn]), extensions[i]) == 0)
            return extensions[i-1];
     }

  return NULL;
}

char *UrlDecode(char *in)
{
  int   i, j, n;
  char *out;
  
  n = strlen(in);
  out = malloc(n + 1);
  
  j=0;
  
  for (i=0; i < n; ++i) 
     {
       if ((i < n - 2) && (in[i] == '%')) /* url encoded as %HH */
         {
           out[j] = 0;
           
           if (isdigit(in[i+1])) 
                 out[j] += 16 * (in[i+1] - '0');
           else if (isalpha(in[i+1])) 
                 out[j] += 16 * (10 + tolower(in[i+1]) - 'a');
                         
           if (isdigit(in[i+2])) 
                 out[j] += in[i+2] - '0';
           else if (isalpha(in[i+2])) 
                 out[j] += 10 + tolower(in[i+2]) - 'a';
                                                   
           i = i + 2;
         }
       else if (in[i] == '+') out[j] = ' ';
       else                   out[j] = in[i];
       
       ++j;
     }

  out[j] = '\0';
  
  return out;
}

char *UrlEncode(char *in)
{
  char *out, *p, *q;
  
  out = malloc(3*strlen(in) + 1);
  
  p = in;
  q = out;
  
  while (*p != '\0')
       {
         if (isalnum(*p) || (*p == '.') || (*p == '=') || (*p == '-'))
           {
             *q = *p;
             ++q;
           }
         else
           {
             sprintf(q, "%%%2x", *p);
             q = &q[3];
           }

         ++p;
       }
  
  *q = '\0';  
  return out;
}

char *MildUrlEncode(char *in)
{
  char *out, *p, *q;
  
  out = malloc(3*strlen(in) + 1);
  
  p = in;
  q = out;
  
  while (*p != '\0')
       {
         if (isalnum(*p) || (*p == '.') || (*p == '=') || (*p == '-') 
                         || (*p == '/') || (*p == '@') || (*p == '_'))
           {
             *q = *p;
             ++q;
           }
         else
           {
             sprintf(q, "%%%2x", *p);
             q = &q[3];
           }

         ++p;
       }
  
  *q = '\0';  
  return out;
}

int GetMemberData(struct grst_member *thisone, char *memberdn)
{
  char *personpath, line[82], *p, *memberdnurl;
  FILE *fp;
 
  thisone->name[0]         = '\0';
  thisone->email[0]        = '\0';
  thisone->organisation[0] = '\0';
  thisone->passcode[0]     = '\0';
  thisone->url[0]          = '\0';
  thisone->addedby[0]      = '\0';
  
  if (memberdn == NULL) return 0;
  
  memberdnurl  = UrlEncode(memberdn);
  personpath = malloc(strlen(GRST_ETCMEMBERS) + strlen(memberdnurl) + 1);
  
  strcpy(personpath, GRST_ETCMEMBERS);
  strcat(personpath, memberdnurl);
 
  fp = fopen(personpath, "r");
  free(personpath);
  if (fp == NULL) return 0; 

  while (fgets(line, sizeof(line), fp) != NULL)
       {
                                                /* 82 = 80 + \n + \0 */
         line[80] = '\0';			/* limit to 80 chars */
         p = index(line, '\n');                
         if (p != NULL) *p = '\0';		/* remove any trailing \n */

         if      (strncmp(line, "name:",     5) == 0)
                strcpy(thisone->name,   &line[5]);
         else if (strncmp(line, "email:",    6) == 0)
                strcpy(thisone->email,  &line[6]);
         else if (strncmp(line, "passcode:", 9) == 0)
                strcpy(thisone->passcode, &line[9]);
         else if (strncmp(line, "organisation:", 13) == 0)
                strcpy(thisone->organisation, &line[13]);
         else if (strncmp(line, "url:",      4) == 0)
                strcpy(thisone->url,    &line[4]);
         else if (strncmp(line, "addedby:",  8) == 0)
                strcpy(thisone->addedby,&line[8]);
       }

  fclose(fp);       
  return 1;
}

int InGroup(char *group, char *dn)
/* returns GRID_ACL_READ if dn[] is in group group[] and
   GRID_ACL_NONE if not (in future, will return GRID_ACL_WRITE etc if
   dn[] should be allowed to modify the group list etc) */
{
  int   i, level = GRID_ACL_NONE;
  char *p, *groupfile, dnline[GRST_ROLELEN+GRST_DNLEN+2], *groupdn;
  FILE *groupfp;

  for (p = group; *p != '\0'; ++p)
    if (!isalnum(*p) && (*p != '@') && (*p != '.') && (*p != ':')) 
      return GRID_ACL_NONE;
  
  groupfile = malloc(strlen(GRST_ETCGROUPS) + 
                     strlen(group) + 1);
         
  strcpy(groupfile, GRST_ETCGROUPS);
  strcat(groupfile, group); 

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: process groupfile %s in InGroup()", 
               getenv("REMOTE_ADDR"), groupfile);
#endif

  groupfp = fopen(groupfile, "r");
  free(groupfile);

  if (groupfp != NULL)
    {
      while (fgets(dnline, GRST_ROLELEN+GRST_DNLEN+2, groupfp) != NULL)
           {
             p = index(dnline, '\n');
             if (p != NULL) *p = '\0';

             p = index(dnline, ' ');
             if (p != NULL)
               {
                 *p = '\0';
                 groupdn = &p[1];

#ifdef GRST_LOGLOCAL
   syslog(GRST_LOGLOCAL + LOG_NOTICE, 
               "%s: in groupfile in InGroup() %s %s", 
               getenv("REMOTE_ADDR"), dnline, groupdn);
#endif
                 if (strcmp(groupdn, dn) == 0)
                   {
                     for (i=0; aclsyms[i] != NULL; ++i)
                      if (strcmp(aclsyms[i], dnline) == 0) 
                        {
                          level = i;
                          break;
                        }
                   }
               }
           }
             
      fclose(groupfp);           
    }         

  return level;
}

int InAcl(char *group)
/* return acl level of group in the ACL of the current directory */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *aclgroup;
  int   i;
  FILE *aclfp;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return GRID_ACL_NONE; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
 
  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp == NULL) return GRID_ACL_NONE;

  while (fgets(oneline, GRST_ROLELEN + GRST_GROUPLEN + 2, aclfp) != NULL)
       {
         p = index(oneline, '\n');
         if (p != NULL) *p = '\0';

         p = index(oneline, ' ');
         if (p != NULL)  
           { 
             *p = '\0';
             aclgroup = &p[1];
           }
         else continue;
         
         if (strcmp(aclgroup, group) == 0)
           for (i=0; aclsyms[i] != NULL; ++i)
              if (strcmp(oneline, aclsyms[i]) == 0)
                {
                  fclose(aclfp);
                  return i;
                } 
       }
    
  fclose(aclfp);
  return GRID_ACL_NONE;
}

void Setacllevel(void)
/* returns GRID_ACL_WRITE, GRID_ACL_READ or GRID_ACL_NONE
   about *userdn based on the ACL in the directory of this page  */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *dn,
       *aclgroup;
  int   level;
  FILE *aclfp;
  
  aclstatus = GRID_ACL_NONE;

  if (userdn == NULL) return;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
      
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: Try to open %s in Setacllevel()", 
               getenv("REMOTE_ADDR"), aclfile);
#endif

  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp != NULL)
    {
      while (fgets(oneline, GRST_ROLELEN + GRST_GROUPLEN + 2, aclfp) != NULL)
       {
         p = index(oneline, '\n');
         if (p != NULL) *p = '\0';

         p = index(oneline, ' ');
         if (p != NULL)  
           { 
             *p = '\0';
             aclgroup = &p[1];
           }
         else continue;

         if (strcmp(aclgroup, GRID_ACL_ANY_DN) == 0)
           {
             level = ConvertLevelSym(oneline);
             if (level > aclstatus) aclstatus = level;
           }
         else if (InGroup(aclgroup, userdn) > GRID_ACL_NONE) 
           {
             level = ConvertLevelSym(oneline);
             if (level > aclstatus) aclstatus = level;
           }
       }

      fclose(aclfp);
    }
#ifdef GRST_LOGLOCAL
  else syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: Cannot open a %s in Setacllevel() %s", 
               getenv("REMOTE_ADDR"), GRST_ACL_FILE, strerror(errno));
#endif
    
  return;
}

void ReadCookies(char **cookiedn, char **passcode)
{
  int   i;
  char *p, *q, *cookies_userdnurl, *cookies_usercode;

  *cookiedn = NULL;
  *passcode = NULL;
  
  p = getenv("HTTP_COOKIE");
  if (p == NULL) return;

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: HTTP cookies offered: %s", getenv("REMOTE_ADDR"), p);
#endif

  cookies_userdnurl = strdup(p);
  cookies_usercode  = strdup(p);

  p = strstr(cookies_userdnurl, "gridsiteDN=");
  if (p != NULL)
    {
      *cookiedn = malloc(strlen(cookies_userdnurl)+1);
      p = &p[11]; /* 11 = strlen("gridsiteDN=") */
      q = *cookiedn;
      while ((*p != ';') && (*p != '\0'))
           {
             if (p[0] == '%') if (isalnum(p[1])) if (isalnum(p[2]))
               {
                 sscanf(&p[1], "%2x", &i);
                 *q = i;
                 ++q;
                 p = &p[3];
                 continue;
               }

             *q = *p;
             ++q;
             ++p;
           }
           
      *q = '\0';
    }
  
  p = strstr(cookies_usercode, "gridsiteCODE=");
  if (p != NULL)
    {
      *passcode = &p[13]; /* 13 = strlen("gridsiteCODE=") */
      while ((*p != ';') && (*p != '\0')) ++p;
      *p = '\0';
    }
}

void DisplayPage(char *rawpage)
{
  int     inblock = 0;
  char   *p, *headtag, *startbody, *foottag, *q;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: Apache (GridSite)\nContent-type: text/html\n");

  headtag = strstr(rawpage, "<body");
  if (headtag == NULL) 
    { 
      headtag = strstr(rawpage, "<BODY");
      if (headtag == NULL) headtag = strstr(rawpage, "<Body");
    }

  if (headtag != NULL)
    {
      for (p = rawpage; p < headtag; ++p) putchar(*p);
      while ((*p != '>') && (*p != '\0')) ++p;
      if (*p == '>') ++p;
      startbody = p;
    }
  else startbody = rawpage;

  CatHeadOrFoot(GRST_HEADFILE);

  foottag = strstr(startbody, "</body>");
  if (foottag == NULL) 
    { 
      foottag = strstr(startbody, "</BODY>");
      if (foottag == NULL) 
        {
          foottag = strstr(startbody, "</Body>");
          if (foottag == NULL) foottag = &startbody[strlen(startbody)];
        }
    }

  puts("<!-- start of user contributed HTML -->");
    
  for (p = startbody; p < foottag; ++p) 
     {
#ifdef GRST_GRIDPP_FMT
       if (*p == '<')
         {
           if (strncasecmp(p, "<h1", 3) == 0)
             {
               if (inblock) 
                   puts("</td></tr></table></center>");

               inblock=1;
               
               fputs("<h1 align=center>", stdout);

               while ((*p != '\0') && (*p != '>')) ++p;
               continue;
             }
           else if (strncasecmp(p, "</h1", 4) == 0)
             {
               puts("</h1>");
               inblock = 0;
                
               while ((*p != '\0') && (*p != '>')) ++p;
               continue;
             }
         }
#endif
         
       if ((!inblock) && (*p != ' ') && (*p != '\t') && (*p != '\n')
           && (*p != '\r'))
         {   
           puts("<center>
              <table width=\"100%\" border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
              <tr><td>");
           inblock = 1;
         }

       putchar(*p);
     }

  puts("<!-- end of user contributed HTML -->");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page body sent", getenv("REMOTE_ADDR"));
#endif

  puts("<p><small>");

  DisplayFileDate();  

  puts("<hr>");
  
  if (userdn != NULL)
    {
      printf("You are %s<br>\n", userdn); 

      if (aclstatus >= GRID_ACL_WRITE)
          if (requesturi != NULL)
           printf("<a href=\"%s?edit\">Edit&nbsp;page</a> .\n", requesturi);

      if (aclstatus >= GRID_ACL_LIST)
          puts("<a href=\"./?list\">List&nbsp;directory</a> .");
    }   

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n",
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");

  free(rawpage);

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListAllMembers()
{
  int        i;
  char      *persondn, **dirarray;
  struct tm *thistm;
  time_t     this_t;
  struct     grst_member thisperson;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: Apache (GridSite)\nContent-type: text/html\n");

  puts("<title>List all website members</title>");
    
  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>List all website members</h1>");

  puts("<table>");
  dirarray = SortDir(GRST_ETCMEMBERS);
  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
         {
           persondn = UrlDecode(dirarray[i]);

           if (GetMemberData(&thisperson, persondn))             
                printf("<tr><td>%s</td>
                        <td><small><a href=\"%s\">%s</a></small></td></tr>\n", 
                       thisperson.name, MildUrlEncode(&persondn[1]), persondn);
           else 
                printf("<tr><td>&nbsp;</td><td><small>%s</small></td></tr>\n",
                       persondn);

           free(persondn);
         }
    }
    
  puts("</table>");

  if ((InGroup(GRST_ADDMBRS,  userdn) > GRID_ACL_NONE) ||
      (InGroup(GRST_EDITMBRS, userdn) > GRID_ACL_NONE))
    printf("<p><form action=\"%s?addmember\" method=post>
            <small><input name=dn size=60>
            <input type=submit value=\"Add\"></small></form>\n",
            GRST_MEMBERS);
          
  puts("<p><small>");
  puts("<hr>");
  
  if (userdn != NULL) printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");


  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void DisplayMember(char *rawpersondn)
{
  int        ngroups = 0, i, editgrps;
  char      *persondn, **dirarray;
  struct tm *thistm;
  time_t     this_t;
  struct grst_member  thisperson;
  
  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }

  persondn = UrlDecode(rawpersondn);

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: Apache (GridSite)\nContent-type: text/html\n");

  printf("<title>%s</title>\n", persondn);

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  if (GetMemberData(&thisperson, persondn))
    {
      puts("<p><table>");
      
      printf("<tr><td>DN:</td><td>%s</td></tr>\n", persondn);

      if (thisperson.name[0] != '\0')
       printf("<tr><td>Name:</td><td>%s</td></tr>\n",  thisperson.name);

      if (thisperson.email[0] != '\0')
       printf("<tr><td>Email:</td><td><a href=\"mailto:%s\">%s</a></td></tr>\n", 
                                         thisperson.email, thisperson.email);

      if (thisperson.organisation[0] != '\0')
       printf("<tr><td>Organisation:</td><td>%s</td></tr>\n", 
                  thisperson.organisation);

      if (thisperson.url[0] != '\0')
       printf("<tr><td>URL:</td><td><a href=\"%s\">%s</a></td></tr>\n", 
                                         thisperson.url, thisperson.url);

      puts("<tr><td>Groups:</td><td>");
      
      ngroups = 0;
      dirarray = SortDir(GRST_ETCGROUPS);
      if (dirarray != NULL)
        {
          for (i=0; dirarray[i] != NULL; ++i)
             if (InGroup(dirarray[i], persondn) > GRID_ACL_NONE)
               {
                 if (ngroups > 0) fputs(", ", stdout);

                 printf("<a href=\"%s%s\">%s</a>", 
                              GRST_GROUPS, dirarray[i], dirarray[i]);
                 ++ngroups;
               }

          free(dirarray);
        }

      if (ngroups == 0) puts("None");
      puts("</td></tr>");

      puts("</table>\n");      
    }
  else printf("<p>No such user: %s\n", persondn);

  /* Let viewer edit this member if they are the member or are in the
     right groups */

  if ((strcmp(userdn, persondn) == 0) ||
      (InGroup(GRST_EDITMBRS, userdn) > GRID_ACL_NONE) ||
      ((InGroup(GRST_ADDMBRS, userdn) > GRID_ACL_NONE) &&
       (strcmp(userdn, thisperson.addedby) == 0)) )
    {
      printf("<form action=\"%s?editmember\" method=post>\n<table>\n", 
              GRST_MEMBERS);

      printf("<input type=hidden name=dn value=\"%s\">\n", persondn);

      printf("<tr><td>Name:</td>
          <td><input name=name value=\"%s\" size=30 maxlength=%d></td></tr>\n",
              thisperson.name, GRST_NAMELEN);

      printf("<tr><td>Email:</td>
          <td><input name=email value=\"%s\" size=30 maxlength=%d></td></tr>\n",
              thisperson.email, GRST_EMAILLEN);

      printf("<tr><td>Organisation:</td>
          <td><input name=organisation value=\"%s\" size=30 maxlength=%d></td></tr>\n",
              thisperson.organisation, GRST_ORGLEN);

      printf("<tr><td>Homepage URL:</td>
          <td><input name=url value=\"%s\" size=40 maxlength=%d></td></tr>\n",
              thisperson.url, GRST_URLLEN);

      puts("<tr><td align=center colspan=2><input type=submit 
                 value=Update></td></tr></table>\n</form>");
    }
  
  /* Let viewer add this person to groups they are admin of */

  editgrps = InGroup(GRST_EDITGRPS, userdn);
  ngroups = 0;      
  dirarray = SortDir(GRST_ETCGROUPS);
  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
         if (InGroup(dirarray[i], persondn) == GRID_ACL_NONE)
            if ((InGroup(dirarray[i], userdn) >= GRID_ACL_WRITE) ||
                (editgrps > GRID_ACL_NONE))
              {
                if (ngroups == 0) 
                 printf("<center><form action=\"%s?editgroup\" method=post><hr>
                         Add member to groups:
                         <select name=group>\n", GRST_GROUPS);

                printf("<option>%s\n", dirarray[i]);

                ++ngroups;
              }

      free(dirarray);
    }

  if (ngroups > 0) 
       printf("</select>
               <input type=hidden name=levelname value=read>
               <input type=hidden name=dn value=\"%s\">
               <input type=submit value=\"Add\"></form></center>",
               persondn);

  puts("<hr><small>");
  
  if (userdn != NULL)
    {
      printf("You are %s<br>\n", userdn); 

      if (aclstatus >= GRID_ACL_WRITE)
          if (requesturi != NULL)
           printf("<a href=\"%s?edit\">Edit&nbsp;details</a> .\n", requesturi);
    }   

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
       
  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n",GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListGroup(char *groupname)
{
  char      *persondn, *p, *groupfile, line[513], *groupnamedup, *dn;
  struct tm *thistm;
  time_t     this_t;
  FILE      *fp;
  struct     grst_member thisperson;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE) aclstatus=GRID_ACL_ADMIN;
  else aclstatus = InGroup(groupname, userdn);
          
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: Apache (GridSite)\nContent-type: text/html\n");

  groupnamedup = strdup(groupname);
  for (p=groupnamedup; (*p != '\0') && isalnum(*p); ++p) ;
  *p = '\0'; /* force end of string, possibly early */

  printf("<title>Group: %s</title>", groupnamedup);
    
  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  printf("<h1>Group: %s</h1>", groupnamedup);

  groupfile = malloc(strlen(GRST_ETCGROUPS) + 2 + strlen(groupnamedup));
  sprintf(groupfile, "%s/%s", GRST_ETCGROUPS, groupnamedup);

  free(groupnamedup);
  
  fp = fopen(groupfile, "r");
  if (fp != NULL)
    {
      puts("<p><table><tr><th>Name</th><th>User DN</th>
                          <th>Access</th></tr>");

      while (fgets(line, sizeof(line), fp) != NULL)
           {
             dn = index(line, ' ');
             if (dn != NULL)
               {
                 *dn = '\0';
                 ++dn;                 
               }
             else dn = "error";
             
             p = index(dn, '\n');
             if (p != NULL) *p = '\0';

             if (aclstatus < GRID_ACL_ADMIN) puts("<tr>"); 
             else printf("<form action=\"%s?editgroup\" method=post><tr>",
                                                              GRST_GROUPS);

             if (GetMemberData(&thisperson, dn))             
                  printf("<td>%s</td>
                          <td><small><a href=\"%s%s\">%s</a></small></td>\n", 
                    thisperson.name, GRST_MEMBERS, MildUrlEncode(&dn[1]), dn);
             else printf("<td>&nbsp;</td><td><small>%s</small></td>\n", dn);

             if (aclstatus < GRID_ACL_ADMIN)
               {
                 if (strcmp(line, aclsyms[GRID_ACL_ADMIN]) == 0)
                      printf("<td align=center>%s</td><td>&nbsp;</td>\n", 
                                   aclsyms[GRID_ACL_ADMIN]);
                 else puts("<td align=center>member</td><td>&nbsp;</td>");
               }
             else
               {
                 puts("<td align=middle><small><select name=levelname>");

                 if (strcmp(line, aclsyms[GRID_ACL_ADMIN]) == 0)
                      printf("<option value=\"%s\" selected>admin
                              <option value=\"%s\">member
                              <option value=\"%s\">remove\n", 
                              aclsyms[GRID_ACL_ADMIN], aclsyms[GRID_ACL_READ],
                                                       aclsyms[GRID_ACL_NONE]);
                 
                 else printf("<option value=\"%s\">admin
                              <option value=\"%s\" selected>member
                              <option value=\"%s\">remove\n", 
                              aclsyms[GRID_ACL_ADMIN], aclsyms[GRID_ACL_READ],
                                                       aclsyms[GRID_ACL_NONE]);
                 
                 printf("<input type=hidden name=group value=\"%s\">
                         <input type=hidden name=dn value=\"%s\">
                         </select>\n", groupname, dn);

                 puts("<input type=submit value=Change>");
                 puts("</small></td>");
               }
             
             if (aclstatus < GRID_ACL_ADMIN) puts("</tr>"); 
             else                            puts("</tr></form>");
           }
       
      fclose(fp);

      puts("</table>\n");
    }
  
  puts("<p><small>");

  puts("<hr>");
  
  if (userdn != NULL) printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListAllGroups(void)
{
  int        i;
  char      *persondn, *dirpath, **dirarray;
  struct tm *thistm;
  time_t     this_t;
  struct     grst_member thisperson;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: Apache (GridSite)\nContent-type: text/html\n");

  puts("<title>List all groups</title>");
    
  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>List all groups</h1>");

  dirarray = SortDir(GRST_ETCGROUPS);

  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
        printf("<p align=center><a href=\"./%s\">%s</a>\n", 
                                       dirarray[i], dirarray[i]);
      free(dirarray);
    }

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE)
    printf("<center><form action=\"%s?addgroup\" method=post>
            <small><input name=group size=%d>
            <input type=submit value=\"Add group\"></small></form></center>\n",
            GRST_MEMBERS, GRST_GROUPLEN);
          
  puts("<p><small>");

  puts("<hr>");
  
  if (userdn != NULL) printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");


  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void EditGroup(void)
{
  int        fd, i;
  char      *p, *levelname, *group, *fullgrouptemp, *fullgroup, oneline[513],
            *groupdn, *dn, *aclaccess;
  struct tm *thistm;
  time_t     this_t;
  FILE      *groupfp, *tmpfp;
    
  time(&this_t);
  thistm = gmtime(&this_t); 

  dn = GetFromCGI("dn");

  group = GetFromCGI("group");

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE) aclstatus=GRID_ACL_ADMIN;
  else aclstatus = InGroup(group, userdn);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that group");
      return;
    }

  levelname = GetFromCGI("levelname");
  for (i=0; aclsyms[i] != NULL; ++i)
    if (strcmp(aclsyms[i], levelname) == 0) break;

  if (aclsyms[i] == NULL)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      printf("Level %s doesnt exist", levelname);
      return;
    }
    
  fullgrouptemp = malloc(sizeof(GRST_ETCGROUPS) + strlen(group) + 9);
  sprintf(fullgrouptemp, "%s/%s.XXXXXX", GRST_ETCGROUPS, group);

  fd = mkstemp(fullgrouptemp);
  if (fd == -1) return;
  close(fd);
  
  tmpfp = fopen(fullgrouptemp, "w");
  if (tmpfp == NULL) 
    {
      free(fullgrouptemp);
      return;
    }

  if (strcmp(levelname, "none") != 0) fprintf(tmpfp, "%s %s\n", levelname, dn);

  fullgroup = malloc(sizeof(GRST_ETCGROUPS) + strlen(group) + 2);
  sprintf(fullgroup, "%s/%s", GRST_ETCGROUPS, group);

  groupfp = fopen(fullgroup, "r");
  if (groupfp == NULL) 
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      printf("Group %s doesnt exist\n", group);
      return;
    }

  while (fgets(oneline, sizeof(oneline), groupfp) != NULL)
           {
             /* now attempt to parse this line: ACCESSLEVEL "DN" */

             aclaccess = oneline;

             while (isspace(*aclaccess)) ++aclaccess;  /* skip leading spaces */
             p = aclaccess;

             while (!isspace(*p) && (*p != '\0')) ++p; /* find end of level */

             if (*p == '\0') continue; /* must be a space: skip line */
         
             *p = '\0';
         
             ++p;
          
             while (isspace(*p)) ++p;
         
             groupdn = p;

             p = index(groupdn, '\n');   /* get rid of tedious trailing \n */
             if (p != NULL) *p = '\0';

             if (*groupdn == '\0') continue; /* DN must exist */

             if (strcasecmp(groupdn, dn) != 0)
                      fprintf(tmpfp, "%s %s\n", aclaccess, groupdn); 
           }

  fclose(groupfp);
  fclose(tmpfp);

  rename(fullgrouptemp, fullgroup);

  free(fullgroup);
  free(fullgrouptemp);
    
  /* everything worked ok so we just redirect to the directory list */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)");

  printf("Location: %s%s\n\n", GRST_GROUPS, group);
}

void EditAcl(void)
{
  int        fd, i;
  char      *p, *levelname, *group, *fullacltemp, *fullacl, oneline[513],
            *aclgroup, *aclaccess;
  struct tm *thistm;
  time_t     this_t;
  FILE      *aclfp, *tmpfp;
    
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_ADMIN)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have admin permission for that ACL");
      return;
    }

  group = GetFromCGI("group");

  levelname = GetFromCGI("levelname");
  for (i=0; aclsyms[i] != NULL; ++i)
    if (strcmp(aclsyms[i], levelname) == 0) break;

  if (aclsyms[i] == NULL)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      printf("Level %s doesnt exist", levelname);
      return;
    }
    
  fullacltemp = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 13);
  
  strcpy(fullacltemp, pathtranslated);
  
  p = rindex(fullacltemp, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(fullacltemp, ".tmp");
  strcat(fullacltemp, GRST_ACL_FILE);
  strcat(fullacltemp, ".XXXXXX");

  fd = mkstemp(fullacltemp);
  if (fd == -1) return;
  close(fd);
  
  tmpfp = fopen(fullacltemp, "w");
  if (tmpfp == NULL) 
    {
      free(fullacltemp);
      return;
    }

  if (strcmp(levelname, "none") != 0) 
       fprintf(tmpfp, "%s %s\n", levelname, group);


  fullacl = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 2);
  strcpy(fullacl, pathtranslated);
  
  p = rindex(fullacl, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(fullacl, GRST_ACL_FILE);

  aclfp = fopen(fullacl, "r");
  if (aclfp != NULL) 
    {    
      while (fgets(oneline, sizeof(oneline), aclfp) != NULL)
           {
             /* now attempt to parse this line: ACCESSLEVEL GROUP */

             aclaccess = oneline;

             while (isspace(*aclaccess)) ++aclaccess;  /* skip leading spaces */
             p = aclaccess;

             while (!isspace(*p) && (*p != '\0')) ++p; /* find end of level */

             if (*p == '\0') continue; /* must be a space: skip line */
         
             *p = '\0';
         
             ++p;
          
             while (isspace(*p)) ++p;
         
             aclgroup = p;

             p = index(aclgroup, '\n'); /* get rid of tedious trailing \n */
             if (p != NULL) *p = '\0';

             if (*aclgroup == '\0') continue; /* group must be given */

             if (strcasecmp(aclgroup, group) != 0)
                      fprintf(tmpfp, "%s %s\n", aclaccess, aclgroup); 
           }
      fclose(aclfp);
    }

  fclose(tmpfp);

  rename(fullacltemp, fullacl);

  free(fullacl);
  free(fullacltemp);
    
  /* everything worked ok so we just redirect to the directory list */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)");

  printf("Location: %s?list#groups\n\n", requestdir);
}

void ConfirmDelete(void)
{
  int        n;
  char      *p;
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>Confirm delete of");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Confirm delete of");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  puts("<p><font size=+5>Do you really want to delete this file?");
         
  fputs("<p align=center><a href=\"", stdout);
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("?reallydelete\">YES!</a>");

  requestdir = strdup(requesturi);
  p = rindex(requestdir, '/');
  p[1] = '\0';
  
  printf("<p align=center><a href=\"%s?list\">NO!</a>\n", requestdir);

  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void DeleteFile(void)
{
  int        n, savederrno;
  char      *p, *filename;
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }
    
  if (remove(pathtranslated) == 0)
    {
      filename = rindex(pathtranslated, '/');
      ++filename;
      UpdateFileLog("Delete", filename, 0, this_t, 0);

      /* everything worked ok so we just redirect to the directory list */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: Apache (GridSite)\nLocation: ", stdout);

      fputs(requestdir, stdout);
      puts("?list");
      return;      
    }

  savederrno = errno;

  puts("HTTP/1.1 500 Denied");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>Cannot delete ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Cannot delete ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>rmdir() failed with error %d: %s\n", 
            savederrno, strerror(savederrno));
         
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void AddGroup(void)
{
  char        *p, *group, *grouppath;
  struct tm   *thistm;
  time_t       this_t;
  struct stat  statbuf;
  FILE        *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (InGroup(GRST_EDITGRPS, userdn) == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have permission to add groups");
      return;
    }
    
  group = GetFromCGI("group");
// check for bad characters
  
  grouppath = malloc(sizeof(GRST_ETCGROUPS) + strlen(group) + 2);
  sprintf(grouppath, "%s/%s", GRST_ETCGROUPS, group);

  if (stat(grouppath, &statbuf) != 0)
    {
      fp = fopen(grouppath, "w");
      if (fp == NULL)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

          puts("Failed to create record for new group");
          return;
        }
      else fclose(fp);
    }

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)");
  printf("Location: %s\n\n", GRST_MEMBERS);
}

void AddMember(void)
{
  char        *p, *dn, *dnencoded, *memberpath;
  struct tm   *thistm;
  time_t       this_t;
  struct stat  statbuf;
  FILE        *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if ((InGroup(GRST_ADDMBRS,  userdn) == GRID_ACL_NONE) &&
      (InGroup(GRST_EDITMBRS, userdn) == GRID_ACL_NONE))
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have permission to add members");
      return;
    }
    
  dn = GetFromCGI("dn");
  dnencoded = UrlEncode(dn);  
  
  memberpath = malloc(sizeof(GRST_ETCMEMBERS) + strlen(dnencoded) + 2);
  sprintf(memberpath, "%s/%s", GRST_ETCMEMBERS, dnencoded);

  if (stat(memberpath, &statbuf) != 0)
    {
      fp = fopen(memberpath, "w");
      if (fp == NULL)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

          puts("Failed to create record for new member");
          return;
        }
      else 
        {
          fprintf(fp, "addedby:%s\n", userdn);
          fclose(fp); 
        }
    }

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)");
  printf("Location: %s%s\n\n", GRST_MEMBERS, MildUrlEncode(&dn[1]));
}

void EditMember(void)
{
  char        *p, *dn, *dnencoded, *memberpath;
  struct tm   *thistm;
  time_t       this_t;
  FILE        *fp;
  struct stat  statbuf;
  struct grst_member thisperson;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  dn = GetFromCGI("dn");

  if (GetMemberData(&thisperson, dn))
    {    
      if ((strcmp(userdn, dn) != 0) &&
          (InGroup(GRST_EDITMBRS, userdn) == GRID_ACL_NONE) &&
          ((InGroup(GRST_ADDMBRS, userdn) == GRID_ACL_NONE) ||
           (strcmp(userdn, thisperson.addedby) != 0)) )
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

          puts("You do not have permission to add members");
          return;
        }

      strncpy(thisperson.name,  GetFromCGI("name"),  GRST_NAMELEN);
      strncpy(thisperson.email, GetFromCGI("email"), GRST_EMAILLEN);
      strncpy(thisperson.url,   GetFromCGI("url"),   GRST_URLLEN);
      strncpy(thisperson.organisation,GetFromCGI("organisation"),GRST_ORGLEN);
      
      dnencoded = UrlEncode(dn);  
      memberpath = malloc(sizeof(GRST_ETCMEMBERS) + strlen(dnencoded) + 2);
      sprintf(memberpath, "%s/%s", GRST_ETCMEMBERS, dnencoded);

      fp = fopen(memberpath, "w");
      if (fp == NULL)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

          printf("Failed to create record for new member %s\n", memberpath);
          return;
        }
      else 
        {
          if (thisperson.name[0] != '\0')
            fprintf(fp, "name:%s\n", thisperson.name);
          if (thisperson.email[0] != '\0')
            fprintf(fp, "email:%s\n", thisperson.email);
          if (thisperson.organisation[0] != '\0')
            fprintf(fp, "organisation:%s\n", thisperson.organisation);
          if (thisperson.passcode[0] != '\0')
            fprintf(fp, "passcode:%s\n", thisperson.passcode);
          if (thisperson.url[0] != '\0')
            fprintf(fp, "url:%s\n", thisperson.url);
          if (thisperson.addedby[0] != '\0')
            fprintf(fp, "addedby:%s\n", thisperson.addedby);

          fclose(fp); 
        }
    }    

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)");
  printf("Location: %s%s\n\n", GRST_MEMBERS, MildUrlEncode(&dn[1]));
}

void CreateDirectory()
{
  int        n, savederrno;
  char      *p, *dirname, *newpathdir, *oldpathfile, *newpathfile;
  struct tm *thistm;
  time_t     this_t;
  size_t     size_dirname;
  FILE      *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }
    
  dirname = GetFromCGI("dirname");

  p = dirname;
  
  while (*p != '\0') 
       {
         if ((*p == '.') || (*p == '/') || (*p == ' ')) 
           {
             puts("HTTP/1.1 500 Denied");
             printf("Date: %s", asctime(thistm));
             puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

             puts("<title>Cannot create ");
             for (p = dirname; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
             puts("</title>");

             CatHeadOrFoot(GRST_HEADFILE);

             puts("<center>
              <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
              <tr><td>");  

             puts("<h1>Cannot create  ");
             for (p = dirname; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
             puts("</h1>");

             printf("<p>Will not create directories with names including
                     . / or SPACE\n");
         
             puts("</td></tr></table></center>");

             puts("</td></tr></table></body>");
             return;
           }
           
         ++p;
       }

  newpathdir = malloc(strlen(docroot) + strlen(requestdir) + 
                      strlen(dirname) + 1);

  strcpy(newpathdir, docroot);
  strcat(newpathdir, requestdir);
  strcat(newpathdir, dirname);
    
  if (mkdir(newpathdir, 0755) == 0)
    {

      oldpathfile = malloc(strlen(docroot) + strlen(requestdir) + 
                           strlen(GRST_ACL_FILE) + 2);
      sprintf(oldpathfile, "%s%s/%s", docroot, requestdir, GRST_ACL_FILE);

      newpathfile = malloc(strlen(newpathdir) + strlen(GRST_ACL_FILE) + 2);
      sprintf(newpathfile, "%s/%s", newpathdir, GRST_ACL_FILE);
      
      CopyFile(newpathfile, oldpathfile);
    
      sprintf(newpathfile, "%s/index.html", newpathdir);
      fp = fopen(newpathfile, "w");
      if (fp != NULL) fclose(fp);

      /* everything worked ok so we just redirect to the directory list */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: Apache (GridSite)\nLocation: ", stdout);

      fputs(requestdir, stdout);
      puts("?list");
      return;      
    }

  savederrno = errno;

  puts("HTTP/1.1 500 Denied");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>Cannot create ");
  for (p = newpathdir; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Cannot create ");
  for (p = newpathdir; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>mkdir() failed with error %d: %s\n", 
            savederrno, strerror(savederrno));
         
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void StatusPage(char *rawcookiedn, char *cookiedn, char *certdn, 
                struct grst_member *thisone)
{
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>GridPP website access status</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>GridPP website access status</h1>");

  printf("<p>(For help with problems, see the <a href=\"%s\">website help
          pages</a>.)\n", GRST_WEBSITE);

  if (ishttps) puts("<p>You are using the Secure (https) server.");
  else         puts("<p>You are using the Insecure (http) server.");

  puts("You can change servers by following the &quot;Switch to ...&quot; 
        link at the bottom of most pages.");
           
  if (userdn == NULL) 
      puts("<p>You have not passed any valid credentials to this site.");
  else
    {
      printf("<p>Your user identity is <b>%s</b>\n", userdn);

      puts("<p>(Cookies: ");
      if (cookiedn != NULL) 
           printf("<b>%s</b>)\n", cookiedn);
      else puts("none)");
     
      puts("<br>(User Cert: ");
      if (certdn != NULL) 
           printf("<b>%s</b>)\n", certdn);
      else puts("none)");

      printf("<p>Contact details:<br><table>
              <tr><td>Name:</td><td>%s</td></tr>
              <tr><td>Email:</td><td>%s</td></tr>
              <tr><td>URL:</td><td>%s</td></tr>
              </table>\n", thisone->name, thisone->email, thisone->url);
    }
    
  puts("<p>You can use this form to obtain cookies for the duration of this
           session (ie until you exit your browser.)");

  puts("<table><form action=\"/?setcookie\" method=post>");
  
  if (rawcookiedn != NULL) 
       printf("<tr><td>DN:</td>
               <td><input type=text name=dn size=40 value=\"%s\"></td></tr>", 
                rawcookiedn);
  else puts(  "<tr><td>DN:</td>
               <td><input type=text name=dn size=40 value=\"/O=Grid/O=UKHEP/OU=\"></td></tr>");
  
  puts("<tr><td>Passcode:</td><td><input type=password width=6 name=passcode></td></tr>");  
  puts("<tr><td colspan=2><input type=submit value=\"Get Cookie\"></td></tr>");
  puts("</table></form>");
      
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void CreationFooter()
{
  int i;

  printf("<form action=\"%s?create\" method=post>\n", requestdir);
  
  puts("<h2>Create a new webpage</h2>");
      
  puts("<table><tr><td>New HTML file:</td>
        <td><input type=text name=filename>
        <input type=hidden name=pagetext 
value=\"&lt;head&gt;
&lt;title&gt;Page title goes here&lt;/title&gt;
&lt;/head&gt;
&lt;body&gt;
&lt;h1&gt;Page heading goes here&lt;/h1&gt;
&lt;p&gt;
A paragraph starts like this and a link to the
&lt;a href=&quot;/&quot;&gt;homepage looks like this&lt;/a&gt;
&lt;/body&gt;\"></td>
        <td><input type=submit value=\"Create\"></td></tr>
        </table></form>");

  printf("<form action=\"%s?createdir\" method=post>\n", requestdir);
  
  puts("<h2>Create a new subdirectory</h2>");
      
  puts("<table><tr><td>New directory:</td>
        <td><input type=text name=dirname></td>
        <td><input type=submit value=\"Create\"></td></tr>
        </table></form>");

  printf("<form action=\"%s?upload\" 
          enctype=\"multipart/form-data\" method=post>\n", requestdir);
  
  puts("<h2>Upload a file / webpage</h2>");

  puts("<p>(The following filetypes are ok: ");
  i=1;
  while (extensions[i] != NULL)
          {
            printf("%s ", extensions[i]);     
            i = i + 2;
          }
  puts(")");

  puts("<table>
        <tr><td>From:</td><td><input type=file name=localfile></td></tr>
        <tr><td>To:</td><td><input type=text name=filename></td></tr>
        <tr><td colspan=2 align=center><input type=submit value=\"Upload\"></td></tr>
        </table></form>");
}

void DisplayAcl(void)
/* Displays Acl in a table for the current directory */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *dn,
       *aclgroup, **dirarray;
  int   level, i, ngroups;
  FILE *aclfp;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
      
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: Try to open %s in DisplayAcl()", 
               getenv("REMOTE_ADDR"), aclfile);
#endif

  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp != NULL)
    {
      puts("<a name=groups><h2>Groups</h2></a>\n<table>");

      while (fgets(oneline, GRST_ROLELEN + GRST_GROUPLEN + 2, aclfp) != NULL)
       {
         p = index(oneline, '\n');
         if (p != NULL) *p = '\0';

         p = index(oneline, ' ');
         if (p != NULL)  
           { 
             *p = '\0';
             aclgroup = &p[1];
           }
         else continue;
         
         if (strcmp(aclgroup, GRID_ACL_ANY_DN) == 0)
              puts("<tr><td>Anyone</td>");
         else printf("<tr><td><a href=\"%s%s\">%s</a></td>\n", 
                              GRST_GROUPS, aclgroup, aclgroup);

         if (aclstatus < GRID_ACL_ADMIN)
             printf("<td>%s</td></tr>\n", oneline);
         else 
           {
             puts("<form action=\"?editacl\" method=post><td align=middle>
                   <small><select name=levelname>");
             
             for (i=0; aclsyms[i] != NULL; ++i)
              if (strcmp(oneline, aclsyms[i]) == 0)
                   printf("<option selected>%s\n", aclsyms[i]);
              else printf("<option>%s\n", aclsyms[i]);
             
             puts("</select>");
             
             printf("<input type=hidden name=group value=\"%s\">\n", aclgroup);
             puts("<input type=submit value=Change></small></td></form></tr>");
           }
       }

      fclose(aclfp);
      puts("</table>");
    }
#ifdef GRST_LOGLOCAL
  else syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: Cannot open a %s in DisplayAcl() %s", 
               getenv("REMOTE_ADDR"), GRST_ACL_FILE, strerror(errno));
#endif
    
  if (aclstatus == GRID_ACL_ADMIN)
    {
      ngroups = 0;      
      dirarray = SortDir(GRST_ETCGROUPS);
      if (dirarray != NULL)
        {
          for (i=0; dirarray[i] != NULL; ++i)
           if (InAcl(dirarray[i]) == GRID_ACL_NONE)
             {
               if (ngroups == 0) 
                 printf("<p><form action=\"%s?editacl\" method=post>
                         <select name=group>\n", requestdir);

               printf("<option>%s\n", dirarray[i]);

               ++ngroups;
             }

          free(dirarray);

          if (ngroups > 0) 
            puts("</select>
                  <input type=hidden name=levelname value=read>
                  <input type=submit value=\"Add to list\"></form>");
        }
    }

  return;
}

void ListDirectory()
{
  int        n;
  char      *p, *q, *dirpath, *filedirpath, *mimetype,
             fileftime[81], *headerpath;
  struct tm *thistm;
  time_t     this_t;
  DIR       *listDIR;
  struct     stat    statbuf;
  struct     dirent *oneentry;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_LIST)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>List directory");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  dirpath = strdup(pathtranslated);

  p = rindex(dirpath, '/');
  if (p != NULL) p[0] = '\0';

  headerpath = malloc(strlen(dirpath) + 12); 
  strcpy(headerpath, requestdir);
  strcat(headerpath, "/HEADER.txt");

  puts("<pre>");
  CatFile(headerpath);
  puts("</pre>");

  puts("<h1>Listing of directory");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  puts("<table>");

  puts("<tr>
        <td colspan=3><a href=\"../\">Parent Directory/</a></td>
        <td colspan=4>&nbsp</td>
        </tr>");

  listDIR = opendir(dirpath);
  if (listDIR != NULL)
        {
          while ((oneentry = readdir(listDIR)) != NULL)
               {
                 if (oneentry->d_name[0] != '.')
                   {
                     filedirpath = malloc(strlen(dirpath) + 2 + 
                                          strlen(oneentry->d_name));
                     sprintf(filedirpath, "%s/%s", dirpath, oneentry->d_name);
                     lstat(filedirpath, &statbuf);
                     free(filedirpath);
                     
                     strftime(fileftime, sizeof(fileftime), 
                              "%R</td><td align=right>%e %b %y",
                              localtime(&statbuf.st_mtime));
                     
                     if (S_ISDIR(statbuf.st_mode)) 
                      printf("<tr><td>&nbsp</td>
                                  <td>&nbsp</td>
                                  <td>&nbsp</td>
                                  <td><a href=\"%s/\">%s/</a></td>
                                  <td>&nbsp</td>
                                  <td align=right>%s</td></tr>",
                            oneentry->d_name, oneentry->d_name, 
                            fileftime);
                     else if ((strcmp(oneentry->d_name, "index.html") != 0)
                          &&  ((strcmp(oneentry->d_name, "HEADER.txt") != 0)
                          ||   (aclstatus >= GRID_ACL_WRITE)))
                       {
                         mimetype = GetMimeType(oneentry->d_name);
                         
                         if (mimetype != NULL)
                           {
                             puts("<tr>");
                             
                             printf("<td><a href=\"%s%s?history\">History</a></td>\n",
                                               requestdir, oneentry->d_name);
                             
                             if (((strcmp(mimetype, "text/html") == 0) ||
                                  (strcmp(mimetype, "text/css")  == 0) ||
                                  (strcmp(mimetype, "text/plain") == 0))
                                 && (aclstatus >= GRID_ACL_WRITE))
                             printf("<td><a href=\"%s%s?edit\">Edit</a></td>\n",
                                    requestdir, oneentry->d_name);
                             else puts("<td>&nbsp;</td>");       

                             if ((strcmp(oneentry->d_name,"index.html") != 0)
                                 && (aclstatus >= GRID_ACL_WRITE))
                              printf("<td><a 
                               href=\"%s%s?confirmdelete\">Delete</a></td>\n",
                               requestdir, oneentry->d_name);
                             else puts("<td>&nbsp;</td>");

                             printf("<td><a href=\"%s\">%s</a></td>
                                     <td align=right>%d</td>
                                     <td align=right>%s</td></tr>",
                                      oneentry->d_name, oneentry->d_name, 
                                      statbuf.st_size,
                                      fileftime);
                           }
                        
                       }
                   }
               }          
          
          closedir(listDIR);
        }

  free(dirpath);
  puts("</table>");

  DisplayAcl();

  if (aclstatus >= GRID_ACL_WRITE) CreationFooter(requestdir);

  puts("<hr><small>");
  
  if (userdn != NULL) printf("You are %s<br>\n", userdn); 
    
  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
     
  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a>\n", GRST_GROUPS);
    
  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n",
             GRST_VERSION);

  puts("</small>");
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
 
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: browse footer sent", getenv("REMOTE_ADDR"));
}

void FileHistory()
{
  int        personret, n, c, filesize, histlen = 0;
  char      *p, *q, *logfilename, *requestdir, *filename, oneline[161], 
             s[81], fileaction[81], *actiondn;
  time_t     this_t, file_t;
  struct tm *thistm, *file_tm;
  FILE      *fp;
  struct     grst_member dnperson;
 
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>History of file ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  printf("<h1>History of file <a href=\"");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  fputs("\">", stdout);
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</a></h1>");

  puts("<table>");
 
  puts("<tr><th>Action</th><th>Date</th><th>File size</th><th>Changed
        by</th></tr>");

  requestdir = strdup(requesturi);
  
  filename = rindex(requestdir, '/');
  *filename = '\0';
  ++filename;
  p = index(filename, '?');
  if (p != NULL) *p = '\0';
  if (*filename == '\0') filename = "index.html";

  logfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 8);

  sprintf(logfilename, "%s/%s/.%s.log", docroot, requestdir, filename);

  fp = fopen(logfilename, "r");
  if (fp != NULL)
    {
      while (fgets(oneline, sizeof(oneline), fp) != NULL)
           {
             ++histlen;
             p = rindex(oneline, '\n');
             if (p != NULL) *p = '\0';
             sscanf(oneline, "%80s %d %ld", fileaction, &filesize, &file_t);
             file_tm = localtime(&file_t);     
             strftime(s, sizeof(s), "%R %e %b %y", file_tm); 
             actiondn = index(oneline, '/');

             personret = GetMemberData(&dnperson, actiondn);
             if (actiondn == NULL) actiondn = "-Not recorded-";
             

             printf("<tr><td>%s</td><td align=right>%s</td>
                         <td align=center>%d</td>\n", 
                         fileaction, s, filesize);

             if (!personret) printf("<td>%s</td>\n", actiondn);
             else            printf("<td><a href=\"%s%s\">%s</td>\n", 
                          GRST_MEMBERS, MildUrlEncode(&actiondn[1]), actiondn);

             puts("</tr>");
           }
     
      fclose(fp);
    }

  free(requestdir);

  if (histlen == 0) 
   puts("<tr><td colspan=4 align=center>No history recorded for this file yet.</td></tr>");

  puts("</table>");

  puts("<hr><small>");
  
  if (userdn != NULL) printf("You are %s<br>\n", userdn); 

  if (aclstatus >= GRID_ACL_WRITE)
      puts("<a href=\"./?list\">List&nbsp;directory</a> .");

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
     
  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");
  puts("</td></tr></table></center>");

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");
}

void EditPage(char *rawpage, char *senderror, char *sendfilename)
{
  char      *p;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

      puts("You do not have write permission to that page/directory");
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>Edit");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Edit");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");
  
  if (senderror != NULL)
      printf("<p><b>Error: %s</b>\n", senderror);

  printf("<form action=\"%s?send\" method=post>\n", requestdir);  
//  printf("<form action=\"/cgi-bin/printenv.cgi\" method=post>\n", requestdir);  

  puts("<p><input type=submit value=\"Save changes\">");

  if (sendfilename != NULL)
      p = sendfilename;
  else
    {
      p = rindex(pathtranslated, '/');
      ++p;
    }
  printf("<input type=text name=filename value=\"%s\">\n", p);

  puts("<p><textarea name=pagetext cols=60 rows=22>");

  for (p=rawpage; *p != '\0'; ++p)
   if      (*p == '<') fputs("&lt;",   stdout);
   else if (*p == '>') fputs("&gt;",   stdout);
   else if (*p == '&') fputs("&amp;",  stdout);
   else if (*p == '"') fputs("&quot;", stdout);
   else putchar(*p);
   
  puts("</textarea>");

  puts("<p><input type=submit value=\"Save changes\">");

  puts("</form>");

  puts("</td></tr></table></center>");

  free(rawpage);

  puts("</td></tr></table></body>");
 
  return;
}

void SetCookie()
{
  char      *dn, *passcode;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  dn       = GetFromCGI("dn");
  passcode = GetFromCGI("passcode");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
         "%s: ?setcookie offers dn=%s passcode=%s", getenv("REMOTE_ADDR"), 
         dn, passcode);
#endif

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  printf("Server: Apache (GridSite)\n");
  printf("Set-Cookie: gridsiteDN=%s; path=/; expires=Wed, 31-Dec-2003 23:59:59 GMT\n",
          dn);
  printf("Set-Cookie: gridsiteCODE=%s; path=/\n",
          passcode);
  printf("Location: %s\n\n", GRST_STATUS);
}

void SendPage()
/* for SendPage, pathtranslated will always be that of the directory, not
   of the file to be written */
{
  int        savederrno = 0, htmltxtflag = 0, n, c;
  char      *p, *pagetext, *filename, *pathfilename, *mimetype,
            *pathfiledatename, *fileext;
  size_t     size_pagetext, size_filename;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }

  filename = GetFromCGI("filename");
  size_filename = strlen(filename) + 1;

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in SendPage(), filename=%s", getenv("REMOTE_ADDR"), 
              filename);
#endif

  fileext = rindex(filename, '.');
  if (fileext == NULL) fileext = "";

  if (index(filename, '/') != NULL)
    {
      pagetext = GetFromCGI("pagetext");

      EditPage(pagetext, "The file name must not contain /", filename);
      free(filename);
      free(pagetext);
      return;            
    }

  mimetype = GetMimeType(filename);

  if (mimetype != NULL)
   if ((strcmp(mimetype, "text/plain") == 0) ||
       (strcmp(mimetype, "text/html")  == 0) ||
       (strcmp(mimetype, "text/css")   == 0))   htmltxtflag = 1;
   
  if (!htmltxtflag)    
    {
      pagetext = GetFromCGI("pagetext");

      EditPage(pagetext, 
         "The file extension must be a text/plain, text/html or text/css type",
          filename);
      free(filename);
      free(pagetext);
      return;            
    }

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in SendPage(), fileext=%s, pathtranslated=%s", 
              getenv("REMOTE_ADDR"), fileext, pathtranslated);
#endif
    
  pathfilename = malloc(strlen(pathtranslated) + 1 + size_filename);
  strcpy(pathfilename, pathtranslated);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in SendPage() after strcpy pathfilename",
              getenv("REMOTE_ADDR"));
#endif

  p = rindex(pathfilename, '/');
  p[1] = '\0';

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in SendPage(), pathfilename1=%s", getenv("REMOTE_ADDR"), 
              pathfilename);
#endif

  pathfiledatename = malloc(strlen(pathfilename) + size_filename + 20);
  sprintf(pathfiledatename, "%s.%s.%ld.%06ld", pathfilename, filename,
           this_t, thisdaytime.tv_usec);

  strcat(pathfilename, filename);
  
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in SendPage(), pathfilename=%s, pathfiledatename=%s", 
              getenv("REMOTE_ADDR"), pathfilename, pathfiledatename);
#endif

  fp = fopen(pathfiledatename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      pagetext = GetFromCGI("pagetext");
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: got pagetext=%s in SendPage()", getenv("REMOTE_ADDR"),
              pagetext);
#endif

      for (p = pagetext; *p != '\0'; ++p) fputc(*p, fp);

      if (fclose(fp) != 0) savederrno = errno;
      else
        {
#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: about to unlink in SendPage()", getenv("REMOTE_ADDR"));
#endif

          unlink(pathfilename);

#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: about to link in SendPage()", getenv("REMOTE_ADDR"));
#endif
          link(pathfiledatename, pathfilename);

          UpdateFileLog("Edit", filename, strlen(pagetext), this_t, 
                                               thisdaytime.tv_usec);
#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: after UpdateFileLog in SendPage()", getenv("REMOTE_ADDR"));
#endif
        }
        
      free(pagetext);
    }

  free(pathfiledatename);

  if (savederrno == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: finished ok in SendPage()", getenv("REMOTE_ADDR"));
#endif

      /* everything worked ok so we just redirect to the new copy */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: Apache (GridSite)\nLocation: ", stdout);

      fputs(requestdir, stdout);
 
      if (strcmp(fileext, ".html") == 0) puts(filename);
      else                               puts("?list");
      
      free(filename);
      return;
    }

  puts("HTTP/1.1 500 Internal error");
  printf("Date: %s", asctime(thistm));
  puts("Server: Apache (GridSite)\nContent-type: text/html\n");      

  puts("<title>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>fopen()/fclose() failed with error %d: %s\n", 
              savederrno, strerror(savederrno));

  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
 
  return;
}

void CreatePage()
{
  int        savederrno = 0;
  char      *p, *pagetext, *filename, *mimetype;
  size_t     size_pagetext, size_filename;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }

  filename = GetFromCGI("filename");
  pagetext = GetFromCGI("pagetext");

  mimetype = GetMimeType(pathtranslated);
  if (mimetype == NULL) mimetype = "application/x-binary";

  if ((index(filename, '/') == NULL) && 
      (strncmp(mimetype, "text/", 5) == 0))
   EditPage(pagetext, NULL, filename);
  else
   EditPage(pagetext, 
    "The filename must be a text format and not contain /",  
      filename);
  
  free(filename);
}

int StrCmpShort(char *long_s, char *short_s)
{
  while (*short_s != '\0')
       {
         if (*long_s == '\0') return -1;         

         if (*long_s > *short_s) return +1;
         if (*long_s < *short_s) return -1;
         
         ++short_s;
         ++long_s;
       }

  return 0;
}

#define MALLINC 1000
#define CONTLEN 200
#define MISTUNKNOWN 1
#define MISTUPLOAD  2
#define MISTFILENM  3
#define MAXFILENAME 1000

void StoreUploadFile(char **filebuffer, int *bufferused, char *boundary)
{
  int bufferlen = 0, c, boundarylen;

  *filebuffer = NULL;
  *bufferused = 0;
  boundarylen = strlen(boundary);

  while ((c = getchar()) != EOF)
       {
         ++(*bufferused);
   
         if (*bufferused > bufferlen)
           {
             bufferlen = bufferlen + MALLINC;
             *filebuffer = realloc(*filebuffer, (size_t) bufferlen);

             if (*filebuffer == NULL) return;
           }

         (*filebuffer)[*bufferused - 1] = c;         
         
         if ( (*bufferused > boundarylen)    &&
              (boundary[boundarylen-1] == c) && 
              (boundary[boundarylen-2] == (*filebuffer)[*bufferused - 2]))
           {
#ifdef GRST_LOGLOCAL
             syslog(GRST_LOGLOCAL + LOG_NOTICE, 
                  "%s: StoreUploadFile buffer %s", 
                  getenv("REMOTE_ADDR"), *filebuffer);
#endif
             if (strncmp(&((*filebuffer)[*bufferused-boundarylen]),
                         boundary, boundarylen) == 0)
               {
                 *bufferused = *bufferused - boundarylen - 2;
                 (*filebuffer)[*bufferused] = '\0';
#ifdef GRST_LOGLOCAL
                 syslog(GRST_LOGLOCAL + LOG_NOTICE, 
                  "%s: StoreUploadFile finds boundary %s and returns buffer %s", 
                  getenv("REMOTE_ADDR"), boundary, *filebuffer);
#endif
                 return;
               }
           }           
       }
       
  (*filebuffer)[*bufferused] = '\0';
       
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
         "%s: StoreUploadFile fails to find boundary", getenv("REMOTE_ADDR"));
#endif

  return;
}

void UploadFile()
{
  char      *contype, *boundary, *p, oneline[200],
             filename[MAXFILENAME] = "", *pathfilename,
            *pathfiledatename, tmpfilename[MAXFILENAME] = "",
            *filebuffer;
  int        mimestate, bufferused = 0, i, savederrno = 0;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }

  contype = strdup(getenv("CONTENT_TYPE"));

  if (StrCmpShort(contype, "multipart/form-data; boundary=") != 0)
    {
      free(contype);
      puts("HTTP/1.1 500 Bad format");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("Wrong mime type");
      return;      
    }  

  boundary = &contype[30];
  mimestate = MISTUNKNOWN;
#ifdef GRST_LOGLOCAL
syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: MISTUNKNOWN", getenv("REMOTE_ADDR"));
#endif

syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: before file read in, boundary %s", 
               getenv("REMOTE_ADDR"), boundary);

  while (fgets(oneline, sizeof(oneline), stdin) != NULL)
     {
#ifdef GRST_LOGLOCAL
       syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: in UploadFile: read -->%s<--", getenv("REMOTE_ADDR"), oneline);
#endif
       if (*oneline == 13) // MIME has CR/LF line breaks, CR=13
         {
           if      (mimestate == MISTUPLOAD)
             {
#ifdef GRST_LOGLOCAL
               syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: call StoreUploadFile()", 
                      getenv("REMOTE_ADDR"));
#endif
               StoreUploadFile(&filebuffer, &bufferused, boundary);
               mimestate = MISTUNKNOWN;
             }
           else if (mimestate == MISTFILENM)
             {
               fgets(tmpfilename, sizeof(tmpfilename), stdin);
               if (*tmpfilename != 13)
                 {
                   p = index(tmpfilename, 13);
                   *p = '\0';
                   if (*tmpfilename != '\0') strcpy(filename, tmpfilename);
                 }
               mimestate = MISTUNKNOWN;
             }
         }
       else if (StrCmpShort(oneline,             
              "Content-Disposition: form-data; name=\"localfile\"; filename=\"")
                == 0)
         {
           mimestate = MISTUPLOAD;
           if (*filename == '\0') 
             {
               strncpy(filename, &oneline[60], MAXFILENAME-1);

               p = rindex(&oneline[60], '\\');
               if (p != NULL) { ++p ; strncpy(filename, p, MAXFILENAME-1); }

               p = rindex(&oneline[60], '/');
               if (p != NULL) { ++p ; strncpy(filename, p, MAXFILENAME-1); }
                                  
               filename[MAXFILENAME-1] = '\0';
               p = index(filename, '"');
               if (p != NULL) *p = '\0'; 
             }
         }
       else if (StrCmpShort(oneline,
                 "Content-Disposition: form-data; name=\"filename\"") == 0)
         {
           mimestate = MISTFILENM;           
         }      
     }

  free(contype);
  
syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: after file read in", getenv("REMOTE_ADDR"));

  if (filebuffer == NULL)
    {
      puts("HTTP/1.1 500 Failed to get upload document");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("Failed to get upload document");
      return;
    }      

  if (filename[0] == '\0')
    {
      puts("HTTP/1.1 500 Something wrong with filename");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("Something wrong with filename: either not given or included /");
      return;
    }      

  if (index(filename, '/') != NULL)
    {
      puts("HTTP/1.1 500 Filename must not include /");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("<p>Filename must not include /<p>You need to go to the right
           directory and do the upload from there.");
      return;
    }      

  if (GetMimeType(filename) == NULL)
    {
      puts("HTTP/1.1 500 Not an acceptable file extension");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("Only file names of the following types / extensions are ok\n<p>");
      
      i = 0;
      while (extensions[i] != NULL)
           {
             printf("%s (%s)<br>", extensions[i], extensions[i+1]);
             i = i + 2;
           }
       
      return;
    }      
 
  pathfilename = malloc(strlen(pathtranslated) + 2 + strlen(filename));

  strcpy(pathfilename, pathtranslated);
  p = rindex(pathfilename, '/');
  p[1] = '\0';

  pathfiledatename = malloc(strlen(pathfilename) + strlen(filename) + 20);
  sprintf(pathfiledatename, "%s.%s.%ld.%06ld", pathfilename, filename, 
          this_t, thisdaytime.tv_usec);
               
  strcat(pathfilename, filename);

  fp = fopen(pathfiledatename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      for (i=0; i < bufferused; ++i) fputc(filebuffer[i], fp);

      if (fclose(fp) != 0) savederrno = errno;
      else
        {
          unlink(pathfilename);
          link(pathfiledatename, pathfilename);

          UpdateFileLog("Upload", filename, bufferused, 
                            this_t, thisdaytime.tv_usec);
        }
   }
	               
  free(pathfilename);

  if (savederrno != 0)
    {
      puts("HTTP/1.1 500 Failed to write out upload document");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n");      
      puts("Failed to write out upload document");
      return;
    }
    
  /* everything worked ok so we just redirect to the new copy */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  fputs("Server: Apache (GridSite)\nLocation: ", stdout);
      
  fputs(requestdir, stdout);
  puts("?list");
}

main()
{
  char  *p, *rawpage, *headtag, *startbody, *foottag, *q,
        *querystring, *mimetype, *cookiedn = NULL, 
        *rawcookiedn = NULL, *passcode = NULL, *certdn = NULL;
  int    i, c, contentlen;
  struct grst_member user_member;
  struct tm *thistm;
  struct stat statbuf;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: connected", getenv("REMOTE_ADDR"));
#endif

  docroot = getenv("DOCUMENT_ROOT");
  requesturi = getenv("REQUEST_URI");
  if ((docroot == NULL) || (requesturi == NULL))
    {
      puts("HTTP/1.1 500 Internal error");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\n\nInternal error"); 
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: empty docroot or requesturi", getenv("REMOTE_ADDR"));
      return;
    }

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: REQUEST_URI %s", 
                                     getenv("REMOTE_ADDR"), requesturi);
#endif

  requestdir = strdup(requesturi);
  p = rindex(requestdir, '/');
  if (p != NULL) p[1] = '\0';
  
  pathtranslated = getenv("PATH_TRANSLATED");
  if (pathtranslated == NULL)
    {
      if (StrCmpShort(requesturi, GRST_MEMBERS) == 0)
        {
          pathtranslated = malloc(strlen(docroot) + strlen(GRST_MEMBERS) + 1);
          strcpy(pathtranslated, docroot);
          strcat(pathtranslated, GRST_MEMBERS);
        }
      else if (StrCmpShort(requesturi, GRST_GROUPS) == 0)
        {
          pathtranslated = malloc(strlen(docroot) + strlen(GRST_GROUPS) + 1);
          strcpy(pathtranslated, docroot);
          strcat(pathtranslated, GRST_GROUPS);
        }
      else
        {      
          puts("HTTP/1.1 404 Not Found");
          printf("Date: %s", asctime(thistm));
          puts("Server: Apache (GridSite)\nContent-type: text/html\n\nPathtranslated not found");      
#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: empty pathtranslated for request %s", 
              getenv("REMOTE_ADDR"), getenv("REQUEST_URI"));
#endif
          return;
        }
    }

  if (getenv("SSL_SERVER_CERT") != NULL) 
    { 
      ishttps = 1;
      certdn = getenv("SSL_CLIENT_S_DN"); /* trivial with certs! */
      userdn = certdn;
      GetMemberData(&user_member, userdn);
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: SSL certdn is %s", getenv("REMOTE_ADDR"), certdn);
 
      ReadCookies(&rawcookiedn, &passcode); /* try from cookies */
      if ((rawcookiedn != NULL) && (passcode != NULL))
        {
          GetMemberData(&user_member, rawcookiedn);

          if (strcmp(passcode, user_member.passcode) == 0) 
            {
              cookiedn = rawcookiedn; /* ie cookie+passcode correct */
              userdn   = rawcookiedn;
            }
        }

      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: after cookies userdn=%s", getenv("REMOTE_ADDR"), userdn);
    }

  if (strcmp(requesturi, GRST_STATUS) == 0)
    {
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_STATUS);
      StatusPage(rawcookiedn, cookiedn, certdn, &user_member);
      return;
    }

  Setacllevel();
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: Setacllevel returns %d", getenv("REMOTE_ADDR"), aclstatus);
                
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 No read permission");
      printf("Date: %s", asctime(thistm));

      CatHeadOrFoot(GRST_HEADFILE);
      printf("<table width=\"100%\" border=1 bgcolor=\"#FFFFFF\" cellpadding=5
       cellspacing=0><tr><td bgcolor=white><h1 align=center>Permission
       denied!</h1>The page you are trying to access is restricted. If you
       believe you should have access, please <a href=\"%s\">check your 
       authentification status</a>.</td></tr></table>\n", GRST_STATUS);      

      puts("<p>");

      if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: permission denied", getenv("REMOTE_ADDR"));
#endif

      return;
    }

  querystring = getenv("QUERY_STRING");
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: querystring %s", getenv("REMOTE_ADDR"), querystring);
#endif
  if (querystring != NULL) 
    {
      if (strcmp(querystring, "upload") == 0) 
        {
          UploadFile();
          return;
        }
    }

  contentlen = 0;
  p = getenv("CONTENT_LENGTH");
  if (p != NULL)
    { 
      sscanf(p, "%d", &contentlen);
      
      cgiposted = malloc(contentlen + 3);
      cgiposted[0] = '&';
      for (i=1; i <= contentlen; ++i)
         {
           c = getchar();
           if (c == EOF) break;
           cgiposted[i] = c;                
         }

      cgiposted[i]   = '&';
      cgiposted[i+1] = '\0';
    }
  else cgiposted = "";
  
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
           "%s: cgiposted[]=%s", getenv("REMOTE_ADDR"), cgiposted);
#endif

  if (querystring != NULL) 
    {
      if (strcmp(querystring, "send") == 0) 
        {
          SendPage();
          return;
        }
      else if (strcmp(querystring, "history") == 0) 
        {
          FileHistory();
          return;
        }
      else if (strcmp(querystring, "create") == 0) 
        {
          CreatePage();
          return;
        }
      else if (strcmp(querystring, "createdir") == 0) 
        {
          CreateDirectory();
          return;
        }
      else if (strcmp(querystring, "confirmdelete") == 0) 
        {
          ConfirmDelete();
          return;
        }
      else if (strcmp(querystring, "reallydelete") == 0) 
        {
          DeleteFile();
          return;
        }
      else if (strcmp(querystring, "list") == 0) 
        {
          ListDirectory();
          return;
        }
      else if (strcmp(querystring, "setcookie") == 0) 
        {
          SetCookie();
          return;
        }
      else if (strcmp(querystring, "editgroup") == 0) 
        {
          EditGroup();
          return;
        }
      else if (strcmp(querystring, "addgroup") == 0) 
        {
          AddGroup();
          return;
        }
      else if (strcmp(querystring, "editacl") == 0) 
        {
          EditAcl();
          return;
        }
      else if (strcmp(querystring, "addmember") == 0) 
        {
          AddMember();
          return;
        }
      else if (strcmp(querystring, "editmember") == 0) 
        {
          EditMember();
          return;
        }
    }
      
  if (StrCmpShort(requesturi, GRST_MEMBERS) == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_MEMBERS);
#endif
      p = &requesturi[strlen(GRST_MEMBERS)-1];

      if (p[1] == '\0')                     ListAllMembers();
      else                                  DisplayMember(p);

      return;
    }

  if (StrCmpShort(requesturi, GRST_GROUPS) == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_GROUPS);
#endif      
      p = &requesturi[strlen(GRST_GROUPS)];

      if (*p == '\0') ListAllGroups();
      else            ListGroup(p);

      return;
    } 

  rawpage = ReadInPathTran(pathtranslated);
  if (rawpage == NULL)
    {
      puts("HTTP/1.1 404 Not Found");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n\nCannot open pathtranslated");      
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: request %s is not found", getenv("REMOTE_ADDR"), 
               getenv("REQUEST_URI"));
      return;
    }

  if (querystring != NULL) if (strcmp(querystring, "edit") == 0) 
    {
      EditPage(rawpage, NULL, NULL);
      return;
    }

  if (querystring != NULL) if (*querystring != '\0') 
    {    
      puts("HTTP/1.1 404 Not Found");
      printf("Date: %s", asctime(thistm));
      printf("Server: Apache (GridSite)\nContent-type: text/html\n\nQuery %s not recognised",
             querystring);
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: query %s is not recognised", getenv("REMOTE_ADDR"), 
               querystring);
      return;
    }
    
  mimetype = GetMimeType(pathtranslated);
  if (mimetype == NULL) mimetype = "text/plain";
  
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: request has mimetype %s", getenv("REMOTE_ADDR"), 
               mimetype);

  if ((strcmp(mimetype, "text/html") == 0) && (*rawpage == '\0'))
       ListDirectory();
  else if (strcmp(mimetype, "text/html") == 0)
       DisplayPage(rawpage);
  else 
    {
      puts("HTTP/1.1 200 OK");
      printf("Date: %s", asctime(thistm));
      printf("Server: Apache (GridSite)\nContent-type: %s\n\n", mimetype);

      fp = fopen(pathtranslated, "r");
      if (fp != NULL) 
        {
          while ((c = fgetc(fp)) != EOF) putchar(c);
          fclose(fp);
        }

     syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: non-HTML mimetype output directly", getenv("REMOTE_ADDR"));
    }
     
  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: exit at end of cgiMain()", getenv("REMOTE_ADDR"));
}
